/*
 * dellemc_s5232f_platform.c - Dell EMC S5232F-C3538 Platform Support.
 *
 * Copyright (c) 2018 Cumulus Networks, Inc.  All Rights Reserved.
 * Author: David Yen (dhyen@cumulusnetworks.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/i2c.h>
#include <linux/gpio.h>
#include <linux/mutex.h>
#include <linux/sysfs.h>
#include <linux/device.h>
#include <linux/hwmon.h>
#include <linux/hwmon-sysfs.h>
#include <linux/i2c-ismt.h>
#include <linux/i2c-mux.h>
#include <linux/i2c-mux-gpio.h>
#include <linux/interrupt.h>
#include <linux/stddef.h>
#include <linux/acpi.h>
#include <linux/io.h>
#include <linux/platform_device.h>
#include <linux/platform_data/at24.h>

#include <linux/cumulus-platform.h>
#include "platform_defs.h"
#include "platform_bitfield.h"

#define DRIVER_NAME		  "dellemc_s5232f_platform"
#define DRIVER_VERSION		  "1.0"

#define ISMT_ADAPTER_NAME	  "SMBus iSMT adapter"
#define I801_ADAPTER_NAME	  "SMBus I801 adapter"

#define PCI_DEVICE_ID_XILINX_FPGA 0x7021
#define FPGA_BLOCK_MAX		  32
#define FPGA_MAX_RETRIES	  5
#define FPGA_BAR		  0
#define FPGA_DESC_ENTRIES	  2
#define FPGA_FAILED_READ_REG	  0xffffffffU

/*
 * This platform has two i2c busses:
 *   SMBus_0: SMBus iSMT adapter at dff9f000
 *   SMBus_1: SMBus I801 adapter at e000
 *
 * SMBus_0 has only one interesting device:
 *   board eeprom (0x50)
 *
 * SMBus_1 has the two JC42 devices:
 *   dram jc42 (0x18)
 *   dram jc42 (0x19)
 *
 * This platform also has a large FPGA on the PCIe bus.	 It is used to
 * access the system LEDs, SFF low speed signals, SFF I2C busses, and
 * miscellaneous control and status registers.
 *
 */

/*
 *
 * SMBus Interface
 *
 */

enum {
	I2C_ISMT_BUS = -1,
	I2C_I801_BUS,
};

mk_eeprom(board, 50, 256, AT24_FLAG_IRUGO);

static struct platform_i2c_device_info i2c_devices[] = {
	mk_i2cdev(I2C_ISMT_BUS, "24c02", 0x50, &board_50_at24),
	mk_i2cdev(I2C_I801_BUS, "jc42",	 0x18, NULL),
	mk_i2cdev(I2C_I801_BUS, "jc42",	 0x19, NULL),
};

/* I2C Initialization */
static void i2c_exit(void)
{
	int i;

	for (i = ARRAY_SIZE(i2c_devices); --i >= 0;) {
		struct i2c_client *c = i2c_devices[i].client;

		if (c) {
			i2c_devices[i].client = NULL;
			i2c_unregister_device(c);
		}
	}
	pr_info("I2C driver unloaded\n");
}

static int i2c_init(void)
{
	int ismt_bus;
	int i801_bus;
	int ret;
	int i;

	ismt_bus = cumulus_i2c_find_adapter(ISMT_ADAPTER_NAME);
	if (ismt_bus < 0) {
		pr_err("could not find the iSMT adapter bus\n");
		ret = -ENODEV;
		goto err_exit;
	}

	i801_bus = cumulus_i2c_find_adapter(I801_ADAPTER_NAME);
	if (i801_bus < 0) {
		pr_err("could not find the i801 adapter bus\n");
		ret = -ENODEV;
		goto err_exit;
	}

	for (i = 0; i < ARRAY_SIZE(i2c_devices); i++) {
		int bus = i2c_devices[i].bus;
		struct i2c_client *client;

		if (bus == I2C_ISMT_BUS)
			bus = ismt_bus;
		else if (bus == I2C_I801_BUS)
			bus = i801_bus;
		client = cumulus_i2c_add_client(bus,
						&i2c_devices[i].board_info);
		if (IS_ERR(client)) {
			ret = PTR_ERR(client);
			goto err_exit;
		}
		i2c_devices[i].client = client;
	}
	pr_info("I2C driver loaded\n");
	return 0;

 err_exit:
	i2c_exit();
	return ret;
}

/*
 *
 * FPGA Interface
 *
 */

/* fpga_id - PCI device ID supported by this driver */
static const struct pci_device_id fpga_id[] = {
	{ PCI_DEVICE(PCI_VENDOR_ID_XILINX, PCI_DEVICE_ID_XILINX_FPGA) },
	{ 0, }
};

MODULE_DEVICE_TABLE(pci, fpga_id);

static struct pci_driver fpga_driver;

/* FPGA Hardware Descriptor */
struct fpga_desc {
	u8 tgtaddr_rw;	/* target address & r/w bit */
	u8 wr_len_cmd;	/* write length in bytes or a command */
	u8 rd_len;	/* read length */
	u8 control;	/* control bits */
	u8 status;	/* status bits */
	u8 retry;	/* collision retry and retry count */
	u8 rxbytes;	/* received bytes */
	u8 txbytes;	/* transmitted bytes */
	u32 dptr_low;	/* lower 32 bit of the data pointer */
	u32 dptr_high;	/* upper 32 bit of the data pointer */
} __packed;

struct fpga_priv {
	u8 __iomem *pbar;                       /* PCIe base address register */
	struct pci_dev *pci_dev;
	struct fpga_desc *hw;			/* descriptor virt base addr */
	dma_addr_t io_rng_dma;			/* descriptor HW base addr */
};

/* Accessor functions for reading and writing the FPGA registers */
static int fpga_read_reg(struct device *dev,
			 int reg,
			 int nregs,
			 u32 *val)
{
	struct fpga_priv *priv = dev_get_drvdata(dev);

	*val = readb(priv->pbar + reg);
	return 0;
}

static int fpga_write_reg(struct device *dev,
			  int reg,
			  int nregs,
			  u32 val)
{
	struct fpga_priv *priv = dev_get_drvdata(dev);

	writeb(val, priv->pbar + reg);
	return 0;
}

/* string arrays to map values into names */
static const char * const led_fan_values[] = {
	"off",
	"amber",
	"green",
	"amber_blinking",
};

static const char * const led_system_values[] = {
	"green_blinking",
	"green",
	"amber",
	"amber_blinking",
};

static const char * const led_beacon_values[] = {
	"off",
	"blue_blinking",
};

static const char * const led_power_values[] = {
	"off",
	"amber",
	"green",
	"amber_blinking",
};

static const char * const led_stack_values[] = {
	"green",
	"off",
};

static const char * const led_blink_values[] = {
	"green",
	"green_blinking",
};

static const char * const led_off_values[] = {
	"off",
	"on",
};

static const char * const led_dot_values[] = {
	"off",
	"on",
};

static const char * const led_digit_values[] = {
	"0",
	"1",
	"2",
	"3",
	"4",
	"5",
	"6",
	"7",
	"8",
	"9",
	"A",
	"B",
	"C",
	"D",
	"E",
	"F",
};

/* Define all the bitfields */
mk_bf_ro(fpga, major_rev,	 0x0000, 8,  8, NULL, 0)
mk_bf_ro(fpga, minor_rev,	 0x0000, 0,  8, NULL, 0)
mk_bf_rw(fpga, scratch,		 0x0004, 0, 32, NULL, 0)
mk_bf_ro(fpga, com_e_type,	 0x0008, 8,  3, NULL, 0)
mk_bf_ro(fpga, board_rev,	 0x0008, 4,  4, NULL, 0)
mk_bf_ro(fpga, board_type,	 0x0008, 0,  4, NULL, 0)
mk_bf_ro(fpga, timestamp1,	 0x000c, 0, 32, NULL, 0)
mk_bf_ro(fpga, timestamp2,	 0x0010, 0, 32, NULL, 0)
mk_bf_ro(fpga, led_fan,		 0x0024, 6,  2, led_fan_values, 0)
mk_bf_rw(fpga, led_system,	 0x0024, 4,  2, led_system_values, 0)
mk_bf_rw(fpga, led_beacon,	 0x0024, 3,  1, led_beacon_values, 0)
mk_bf_rw(fpga, led_power,	 0x0024, 1,  2, led_power_values, 0)
mk_bf_ro(fpga, led_stack,	 0x0024, 0,  1, led_stack_values, 0)
mk_bf_rw(fpga, led_blink,	 0x0024, 6,  1, led_blink_values, 0)
mk_bf_rw(fpga, led_off,		 0x0028, 5,  1, led_off_values, 0)
mk_bf_rw(fpga, led_dot,		 0x0028, 4,  1, led_dot_values, 0)
mk_bf_rw(fpga, led_digit,	 0x0028, 0,  4, led_digit_values, 0)
mk_bf_ro(fpga, psu_pwr2_all_ok,	 0x0038, 3,  1, NULL, 0)
mk_bf_ro(fpga, psu_pwr2_present, 0x0038, 2,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, psu_pwr1_all_ok,	 0x0038, 1,  1, NULL, 0)
mk_bf_ro(fpga, psu_pwr1_present, 0x0038, 0,  1, NULL, BF_COMPLEMENT)

/* plumb all of the front panel port signals */
mk_bf_rw(fpga, port1_lpmode,	 0x4000, 6,  1, NULL, 0)
mk_bf_rw(fpga, port1_modsel,	 0x4000, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port1_reset,	 0x4000, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port1_present,	 0x4004, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port2_lpmode,	 0x4010, 6,  1, NULL, 0)
mk_bf_rw(fpga, port2_modsel,	 0x4010, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port2_reset,	 0x4010, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port2_present,	 0x4014, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port3_lpmode,	 0x4020, 6,  1, NULL, 0)
mk_bf_rw(fpga, port3_modsel,	 0x4020, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port3_reset,	 0x4020, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port3_present,	 0x4024, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port4_lpmode,	 0x4030, 6,  1, NULL, 0)
mk_bf_rw(fpga, port4_modsel,	 0x4030, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port4_reset,	 0x4030, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port4_present,	 0x4034, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port5_lpmode,	 0x4040, 6,  1, NULL, 0)
mk_bf_rw(fpga, port5_modsel,	 0x4040, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port5_reset,	 0x4040, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port5_present,	 0x4044, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port6_lpmode,	 0x4050, 6,  1, NULL, 0)
mk_bf_rw(fpga, port6_modsel,	 0x4050, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port6_reset,	 0x4050, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port6_present,	 0x4054, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port7_lpmode,	 0x4060, 6,  1, NULL, 0)
mk_bf_rw(fpga, port7_modsel,	 0x4060, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port7_reset,	 0x4060, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port7_present,	 0x4064, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port8_lpmode,	 0x4070, 6,  1, NULL, 0)
mk_bf_rw(fpga, port8_modsel,	 0x4070, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port8_reset,	 0x4070, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port8_present,	 0x4074, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port9_lpmode,	 0x4080, 6,  1, NULL, 0)
mk_bf_rw(fpga, port9_modsel,	 0x4080, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port9_reset,	 0x4080, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port9_present,	 0x4084, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port10_lpmode,    0x4090, 6,  1, NULL, 0)
mk_bf_rw(fpga, port10_modsel,	 0x4090, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port10_reset,	 0x4090, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port10_present,	 0x4094, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port11_lpmode,    0x40a0, 6,  1, NULL, 0)
mk_bf_rw(fpga, port11_modsel,	 0x40a0, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port11_reset,	 0x40a0, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port11_present,	 0x40a4, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port12_lpmode,    0x40b0, 6,  1, NULL, 0)
mk_bf_rw(fpga, port12_modsel,	 0x40b0, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port12_reset,	 0x40b0, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port12_present,	 0x40b4, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port13_lpmode,    0x40c0, 6,  1, NULL, 0)
mk_bf_rw(fpga, port13_modsel,	 0x40c0, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port13_reset,	 0x40c0, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port13_present,	 0x40c4, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port14_lpmode,    0x40d0, 6,  1, NULL, 0)
mk_bf_rw(fpga, port14_modsel,	 0x40d0, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port14_reset,	 0x40d0, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port14_present,	 0x40d4, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port15_lpmode,    0x40e0, 6,  1, NULL, 0)
mk_bf_rw(fpga, port15_modsel,	 0x40e0, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port15_reset,	 0x40e0, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port15_present,	 0x40e4, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port16_lpmode,    0x40f0, 6,  1, NULL, 0)
mk_bf_rw(fpga, port16_modsel,	 0x40f0, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port16_reset,	 0x40f0, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port16_present,	 0x40f4, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port17_lpmode,    0x4100, 6,  1, NULL, 0)
mk_bf_rw(fpga, port17_modsel,	 0x4100, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port17_reset,	 0x4100, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port17_present,	 0x4104, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port18_lpmode,    0x4110, 6,  1, NULL, 0)
mk_bf_rw(fpga, port18_modsel,	 0x4110, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port18_reset,	 0x4110, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port18_present,	 0x4114, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port19_lpmode,    0x4120, 6,  1, NULL, 0)
mk_bf_rw(fpga, port19_modsel,	 0x4120, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port19_reset,	 0x4120, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port19_present,	 0x4124, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port20_lpmode,    0x4130, 6,  1, NULL, 0)
mk_bf_rw(fpga, port20_modsel,	 0x4130, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port20_reset,	 0x4130, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port20_present,	 0x4134, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port21_lpmode,    0x4140, 6,  1, NULL, 0)
mk_bf_rw(fpga, port21_modsel,	 0x4140, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port21_reset,	 0x4140, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port21_present,	 0x4144, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port22_lpmode,    0x4150, 6,  1, NULL, 0)
mk_bf_rw(fpga, port22_modsel,	 0x4150, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port22_reset,	 0x4150, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port22_present,	 0x4154, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port23_lpmode,    0x4160, 6,  1, NULL, 0)
mk_bf_rw(fpga, port23_modsel,	 0x4160, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port23_reset,	 0x4160, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port23_present,	 0x4164, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port24_lpmode,    0x4170, 6,  1, NULL, 0)
mk_bf_rw(fpga, port24_modsel,	 0x4170, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port24_reset,	 0x4170, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port24_present,	 0x4174, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port25_lpmode,    0x4180, 6,  1, NULL, 0)
mk_bf_rw(fpga, port25_modsel,	 0x4180, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port25_reset,	 0x4180, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port25_present,	 0x4184, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port26_lpmode,    0x4190, 6,  1, NULL, 0)
mk_bf_rw(fpga, port26_modsel,	 0x4190, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port26_reset,	 0x4190, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port26_present,	 0x4194, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port27_lpmode,    0x41a0, 6,  1, NULL, 0)
mk_bf_rw(fpga, port27_modsel,	 0x41a0, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port27_reset,	 0x41a0, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port27_present,	 0x41a4, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port28_lpmode,    0x41b0, 6,  1, NULL, 0)
mk_bf_rw(fpga, port28_modsel,	 0x41b0, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port28_reset,	 0x41b0, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port28_present,	 0x41b4, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port29_lpmode,    0x41c0, 6,  1, NULL, 0)
mk_bf_rw(fpga, port29_modsel,	 0x41c0, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port29_reset,	 0x41c0, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port29_present,	 0x41c4, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port30_lpmode,    0x41d0, 6,  1, NULL, 0)
mk_bf_rw(fpga, port30_modsel,	 0x41d0, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port30_reset,	 0x41d0, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port30_present,	 0x41d4, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port31_lpmode,    0x41e0, 6,  1, NULL, 0)
mk_bf_rw(fpga, port31_modsel,	 0x41e0, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port31_reset,	 0x41e0, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port31_present,	 0x41e4, 4,  1, NULL, BF_COMPLEMENT)

mk_bf_rw(fpga, port32_lpmode,    0x41f0, 6,  1, NULL, 0)
mk_bf_rw(fpga, port32_modsel,	 0x41f0, 5,  1, NULL, BF_COMPLEMENT)
mk_bf_rw(fpga, port32_reset,	 0x41f0, 4,  1, NULL, BF_COMPLEMENT)
mk_bf_ro(fpga, port32_present,	 0x41f4, 4,  1, NULL, BF_COMPLEMENT)

/* I2C interface to pluggable modules */
mk_bf_rw(fpga, i2c1_prerlo,	 0x6000, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c1_prerhi,	 0x6001, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c1_control,	 0x6002, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c1_transmit,	 0x6003, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c1_receive,	 0x6003, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c1_command,	 0x6004, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c1_status,	 0x6004, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c2_prerlo,	 0x6010, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c2_prerhi,	 0x6011, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c2_control,	 0x6012, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c2_transmit,	 0x6013, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c2_receive,	 0x6013, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c2_command,	 0x6014, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c2_status,	 0x6014, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c3_prerlo,	 0x6020, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c3_prerhi,	 0x6021, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c3_control,	 0x6022, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c3_transmit,	 0x6023, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c3_receive,	 0x6023, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c3_command,	 0x6024, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c3_status,	 0x6024, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c4_prerlo,	 0x6030, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c4_prerhi,	 0x6031, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c4_control,	 0x6032, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c4_transmit,	 0x6033, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c4_receive,	 0x6033, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c4_command,	 0x6034, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c4_status,	 0x6034, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c5_prerlo,	 0x6040, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c5_prerhi,	 0x6041, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c5_control,	 0x6042, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c5_transmit,	 0x6043, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c5_receive,	 0x6043, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c5_command,	 0x6044, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c5_status,	 0x6044, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c6_prerlo,	 0x6050, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c6_prerhi,	 0x6051, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c6_control,	 0x6052, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c6_transmit,	 0x6053, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c6_receive,	 0x6053, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c6_command,	 0x6054, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c6_status,	 0x6054, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c7_prerlo,	 0x6060, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c7_prerhi,	 0x6061, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c7_control,	 0x6062, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c7_transmit,	 0x6063, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c7_receive,	 0x6063, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c7_command,	 0x6064, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c7_status,	 0x6064, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c8_prerlo,	 0x6070, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c8_prerhi,	 0x6071, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c8_control,	 0x6072, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c8_transmit,	 0x6073, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c8_receive,	 0x6073, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c8_command,	 0x6074, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c8_status,	 0x6074, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c9_prerlo,	 0x6080, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c9_prerhi,	 0x6081, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c9_control,	 0x6082, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c9_transmit,	 0x6083, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c9_receive,	 0x6083, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c9_command,	 0x6084, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c9_status,	 0x6084, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c10_prerlo,	 0x6090, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c10_prerhi,	 0x6091, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c10_control,	 0x6092, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c10_transmit,	 0x6093, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c10_receive,	 0x6093, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c10_command,	 0x6094, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c10_status,	 0x6094, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c11_prerlo,	 0x60a0, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c11_prerhi,	 0x60a1, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c11_control,	 0x60a2, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c11_transmit,	 0x60a3, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c11_receive,	 0x60a3, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c11_command,	 0x60a4, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c11_status,	 0x60a4, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c12_prerlo,	 0x60b0, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c12_prerhi,	 0x60b1, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c12_control,	 0x60b2, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c12_transmit,	 0x60b3, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c12_receive,	 0x60b3, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c12_command,	 0x60b4, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c12_status,	 0x60b4, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c13_prerlo,	 0x60c0, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c13_prerhi,	 0x60c1, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c13_control,	 0x60c2, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c13_transmit,	 0x60c3, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c13_receive,	 0x60c3, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c13_command,	 0x60c4, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c13_status,	 0x60c4, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c14_prerlo,	 0x60d0, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c14_prerhi,	 0x60d1, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c14_control,	 0x60d2, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c14_transmit,	 0x60d3, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c14_receive,	 0x60d3, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c14_command,	 0x60d4, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c14_status,	 0x60d4, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c15_prerlo,	 0x60e0, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c15_prerhi,	 0x60e1, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c15_control,	 0x60e2, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c15_transmit,	 0x60e3, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c15_receive,	 0x60e3, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c15_command,	 0x60e4, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c15_status,	 0x60e4, 0,  8, NULL, 0)

mk_bf_rw(fpga, i2c16_prerlo,	 0x60f0, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c16_prerhi,	 0x60f1, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c16_control,	 0x60f2, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c16_transmit,	 0x60f3, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c16_receive,	 0x60f3, 0,  8, NULL, 0)
mk_bf_rw(fpga, i2c16_command,	 0x60f4, 0,  8, NULL, 0)
mk_bf_ro(fpga, i2c16_status,	 0x60f4, 0,  8, NULL, 0)

static struct attribute *fpga_attrs[] = {
	&fpga_major_rev.attr,
	&fpga_minor_rev.attr,
	&fpga_scratch.attr,
	&fpga_com_e_type.attr,
	&fpga_board_rev.attr,
	&fpga_board_type.attr,
	&fpga_timestamp1.attr,
	&fpga_timestamp2.attr,
	&fpga_led_fan.attr,
	&fpga_led_system.attr,
	&fpga_led_beacon.attr,
	&fpga_led_power.attr,
	&fpga_led_stack.attr,
	&fpga_led_blink.attr,
	&fpga_led_off.attr,
	&fpga_led_dot.attr,
	&fpga_led_digit.attr,
	&fpga_psu_pwr2_all_ok.attr,
	&fpga_psu_pwr2_present.attr,
	&fpga_psu_pwr1_all_ok.attr,
	&fpga_psu_pwr1_present.attr,

	&fpga_port1_lpmode.attr,
	&fpga_port1_modsel.attr,
	&fpga_port1_reset.attr,
	&fpga_port1_present.attr,
	&fpga_port2_lpmode.attr,
	&fpga_port2_modsel.attr,
	&fpga_port2_reset.attr,
	&fpga_port2_present.attr,
	&fpga_port3_lpmode.attr,
	&fpga_port3_modsel.attr,
	&fpga_port3_reset.attr,
	&fpga_port3_present.attr,
	&fpga_port4_lpmode.attr,
	&fpga_port4_modsel.attr,
	&fpga_port4_reset.attr,
	&fpga_port4_present.attr,
	&fpga_port5_lpmode.attr,
	&fpga_port5_modsel.attr,
	&fpga_port5_reset.attr,
	&fpga_port5_present.attr,
	&fpga_port6_lpmode.attr,
	&fpga_port6_modsel.attr,
	&fpga_port6_reset.attr,
	&fpga_port6_present.attr,
	&fpga_port7_lpmode.attr,
	&fpga_port7_modsel.attr,
	&fpga_port7_reset.attr,
	&fpga_port7_present.attr,
	&fpga_port8_lpmode.attr,
	&fpga_port8_modsel.attr,
	&fpga_port8_reset.attr,
	&fpga_port8_present.attr,
	&fpga_port9_lpmode.attr,
	&fpga_port9_modsel.attr,
	&fpga_port9_reset.attr,
	&fpga_port9_present.attr,
	&fpga_port10_lpmode.attr,
	&fpga_port10_modsel.attr,
	&fpga_port10_reset.attr,
	&fpga_port10_present.attr,
	&fpga_port11_lpmode.attr,
	&fpga_port11_modsel.attr,
	&fpga_port11_reset.attr,
	&fpga_port11_present.attr,
	&fpga_port12_lpmode.attr,
	&fpga_port12_modsel.attr,
	&fpga_port12_reset.attr,
	&fpga_port12_present.attr,
	&fpga_port13_lpmode.attr,
	&fpga_port13_modsel.attr,
	&fpga_port13_reset.attr,
	&fpga_port13_present.attr,
	&fpga_port14_lpmode.attr,
	&fpga_port14_modsel.attr,
	&fpga_port14_reset.attr,
	&fpga_port14_present.attr,
	&fpga_port15_lpmode.attr,
	&fpga_port15_modsel.attr,
	&fpga_port15_reset.attr,
	&fpga_port15_present.attr,
	&fpga_port16_lpmode.attr,
	&fpga_port16_modsel.attr,
	&fpga_port16_reset.attr,
	&fpga_port16_present.attr,
	&fpga_port17_lpmode.attr,
	&fpga_port17_modsel.attr,
	&fpga_port17_reset.attr,
	&fpga_port17_present.attr,
	&fpga_port18_lpmode.attr,
	&fpga_port18_modsel.attr,
	&fpga_port18_reset.attr,
	&fpga_port18_present.attr,
	&fpga_port19_lpmode.attr,
	&fpga_port19_modsel.attr,
	&fpga_port19_reset.attr,
	&fpga_port19_present.attr,
	&fpga_port20_lpmode.attr,
	&fpga_port20_modsel.attr,
	&fpga_port20_reset.attr,
	&fpga_port20_present.attr,
	&fpga_port21_lpmode.attr,
	&fpga_port21_modsel.attr,
	&fpga_port21_reset.attr,
	&fpga_port21_present.attr,
	&fpga_port22_lpmode.attr,
	&fpga_port22_modsel.attr,
	&fpga_port22_reset.attr,
	&fpga_port22_present.attr,
	&fpga_port23_lpmode.attr,
	&fpga_port23_modsel.attr,
	&fpga_port23_reset.attr,
	&fpga_port23_present.attr,
	&fpga_port24_lpmode.attr,
	&fpga_port24_modsel.attr,
	&fpga_port24_reset.attr,
	&fpga_port24_present.attr,
	&fpga_port25_lpmode.attr,
	&fpga_port25_modsel.attr,
	&fpga_port25_reset.attr,
	&fpga_port25_present.attr,
	&fpga_port26_lpmode.attr,
	&fpga_port26_modsel.attr,
	&fpga_port26_reset.attr,
	&fpga_port26_present.attr,
	&fpga_port27_lpmode.attr,
	&fpga_port27_modsel.attr,
	&fpga_port27_reset.attr,
	&fpga_port27_present.attr,
	&fpga_port28_lpmode.attr,
	&fpga_port28_modsel.attr,
	&fpga_port28_reset.attr,
	&fpga_port28_present.attr,
	&fpga_port29_lpmode.attr,
	&fpga_port29_modsel.attr,
	&fpga_port29_reset.attr,
	&fpga_port29_present.attr,
	&fpga_port30_lpmode.attr,
	&fpga_port30_modsel.attr,
	&fpga_port30_reset.attr,
	&fpga_port30_present.attr,
	&fpga_port31_lpmode.attr,
	&fpga_port31_modsel.attr,
	&fpga_port31_reset.attr,
	&fpga_port31_present.attr,
	&fpga_port32_lpmode.attr,
	&fpga_port32_modsel.attr,
	&fpga_port32_reset.attr,
	&fpga_port32_present.attr,

	&fpga_i2c1_prerlo.attr,
	&fpga_i2c1_prerhi.attr,
	&fpga_i2c1_control.attr,
	&fpga_i2c1_transmit.attr,
	&fpga_i2c1_receive.attr,
	&fpga_i2c1_command.attr,
	&fpga_i2c1_status.attr,

	&fpga_i2c2_prerlo.attr,
	&fpga_i2c2_prerhi.attr,
	&fpga_i2c2_control.attr,
	&fpga_i2c2_transmit.attr,
	&fpga_i2c2_receive.attr,
	&fpga_i2c2_command.attr,
	&fpga_i2c2_status.attr,

	&fpga_i2c3_prerlo.attr,
	&fpga_i2c3_prerhi.attr,
	&fpga_i2c3_control.attr,
	&fpga_i2c3_transmit.attr,
	&fpga_i2c3_receive.attr,
	&fpga_i2c3_command.attr,
	&fpga_i2c3_status.attr,

	&fpga_i2c4_prerlo.attr,
	&fpga_i2c4_prerhi.attr,
	&fpga_i2c4_control.attr,
	&fpga_i2c4_transmit.attr,
	&fpga_i2c4_receive.attr,
	&fpga_i2c4_command.attr,
	&fpga_i2c4_status.attr,

	&fpga_i2c5_prerlo.attr,
	&fpga_i2c5_prerhi.attr,
	&fpga_i2c5_control.attr,
	&fpga_i2c5_transmit.attr,
	&fpga_i2c5_receive.attr,
	&fpga_i2c5_command.attr,
	&fpga_i2c5_status.attr,

	&fpga_i2c6_prerlo.attr,
	&fpga_i2c6_prerhi.attr,
	&fpga_i2c6_control.attr,
	&fpga_i2c6_transmit.attr,
	&fpga_i2c6_receive.attr,
	&fpga_i2c6_command.attr,
	&fpga_i2c6_status.attr,

	&fpga_i2c7_prerlo.attr,
	&fpga_i2c7_prerhi.attr,
	&fpga_i2c7_control.attr,
	&fpga_i2c7_transmit.attr,
	&fpga_i2c7_receive.attr,
	&fpga_i2c7_command.attr,
	&fpga_i2c7_status.attr,

	&fpga_i2c8_prerlo.attr,
	&fpga_i2c8_prerhi.attr,
	&fpga_i2c8_control.attr,
	&fpga_i2c8_transmit.attr,
	&fpga_i2c8_receive.attr,
	&fpga_i2c8_command.attr,
	&fpga_i2c8_status.attr,

	&fpga_i2c9_prerlo.attr,
	&fpga_i2c9_prerhi.attr,
	&fpga_i2c9_control.attr,
	&fpga_i2c9_transmit.attr,
	&fpga_i2c9_receive.attr,
	&fpga_i2c9_command.attr,
	&fpga_i2c9_status.attr,

	&fpga_i2c10_prerlo.attr,
	&fpga_i2c10_prerhi.attr,
	&fpga_i2c10_control.attr,
	&fpga_i2c10_transmit.attr,
	&fpga_i2c10_receive.attr,
	&fpga_i2c10_command.attr,
	&fpga_i2c10_status.attr,

	&fpga_i2c11_prerlo.attr,
	&fpga_i2c11_prerhi.attr,
	&fpga_i2c11_control.attr,
	&fpga_i2c11_transmit.attr,
	&fpga_i2c11_receive.attr,
	&fpga_i2c11_command.attr,
	&fpga_i2c11_status.attr,

	&fpga_i2c12_prerlo.attr,
	&fpga_i2c12_prerhi.attr,
	&fpga_i2c12_control.attr,
	&fpga_i2c12_transmit.attr,
	&fpga_i2c12_receive.attr,
	&fpga_i2c12_command.attr,
	&fpga_i2c12_status.attr,

	&fpga_i2c13_prerlo.attr,
	&fpga_i2c13_prerhi.attr,
	&fpga_i2c13_control.attr,
	&fpga_i2c13_transmit.attr,
	&fpga_i2c13_receive.attr,
	&fpga_i2c13_command.attr,
	&fpga_i2c13_status.attr,

	&fpga_i2c14_prerlo.attr,
	&fpga_i2c14_prerhi.attr,
	&fpga_i2c14_control.attr,
	&fpga_i2c14_transmit.attr,
	&fpga_i2c14_receive.attr,
	&fpga_i2c14_command.attr,
	&fpga_i2c14_status.attr,

	&fpga_i2c15_prerlo.attr,
	&fpga_i2c15_prerhi.attr,
	&fpga_i2c15_control.attr,
	&fpga_i2c15_transmit.attr,
	&fpga_i2c15_receive.attr,
	&fpga_i2c15_command.attr,
	&fpga_i2c15_status.attr,

	&fpga_i2c16_prerlo.attr,
	&fpga_i2c16_prerhi.attr,
	&fpga_i2c16_control.attr,
	&fpga_i2c16_transmit.attr,
	&fpga_i2c16_receive.attr,
	&fpga_i2c16_command.attr,
	&fpga_i2c16_status.attr,

	NULL,
};

static struct attribute_group  fpga_attr_group = {
	.attrs = fpga_attrs,
};

static int fpga_dev_init(struct fpga_priv *priv)
{
	/* allocate memory for the descriptor */
	priv->hw = dmam_alloc_coherent(&priv->pci_dev->dev,
				       (FPGA_DESC_ENTRIES
					* sizeof(struct fpga_desc)),
				       &priv->io_rng_dma,
				       GFP_KERNEL);
	if (!priv->hw)
		return -ENOMEM;

	memset(priv->hw, 0, (FPGA_DESC_ENTRIES * sizeof(struct fpga_desc)));

	return 0;
}

static int
fpga_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
	int err;
	struct fpga_priv *priv;
	unsigned long start, len;

	priv = devm_kzalloc(&pdev->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	pci_set_drvdata(pdev, priv);

	priv->pci_dev = pdev;

	err = pcim_enable_device(pdev);
	if (err) {
		dev_err(&pdev->dev, "Failed to enable FPGA PCI device (%d)\n",
			err);
		return err;
	}

	/* enable bus mastering */
	pci_set_master(pdev);

	/* Determine the address of the FPGA area */
	start = pci_resource_start(pdev, FPGA_BAR);
	len = pci_resource_len(pdev, FPGA_BAR);
	if (!start || !len) {
		dev_err(&pdev->dev,
			"FPGA base address uninitialized, upgrade BIOS\n");
		return -ENODEV;
	}

	dev_dbg(&priv->pci_dev->dev, " start=0x%lX\n", start);
	dev_dbg(&priv->pci_dev->dev, " len=0x%lX\n", len);

	err = pci_request_region(pdev, FPGA_BAR, fpga_driver.name);
	if (err) {
		dev_err(&pdev->dev,
			"Failed to request FPGA memory region 0x%lx-0x%lx\n",
			start, start + len);
		return err;
	}
	priv->pbar = pcim_iomap(pdev, FPGA_BAR, len);
	if (!priv->pbar) {
		dev_err(&pdev->dev, "Unable to ioremap FPGA BAR\n");
		err = -ENODEV;
		goto fail;
	}

	if ((pci_set_dma_mask(pdev, DMA_BIT_MASK(64)) != 0) ||
	    (pci_set_consistent_dma_mask(pdev, DMA_BIT_MASK(64)) != 0)) {
		if ((pci_set_dma_mask(pdev, DMA_BIT_MASK(32)) != 0) ||
		    (pci_set_consistent_dma_mask(pdev,
						 DMA_BIT_MASK(32)) != 0)) {
			dev_err(&pdev->dev, "pci_set_dma_mask fail %p\n",
				pdev);
			err = -ENODEV;
			goto fail;
		}
	}

	err = fpga_dev_init(priv);
	if (err)
		goto fail;

	priv->pbar = ioremap(start, len);

	err = sysfs_create_group(&pdev->dev.kobj, &fpga_attr_group);
	if (err) {
		pr_err("sysfs_fpga_attr_group failed for FPGA driver\n");
		goto fail;
	}

fail:
	pci_release_region(pdev, FPGA_BAR);
	return err;
}

static void fpga_remove(struct pci_dev *pdev)
{
	struct kobject *kobj = &pdev->dev.kobj;

	sysfs_remove_group(kobj, &fpga_attr_group);
	pci_release_region(pdev, FPGA_BAR);
}

#define fpga_suspend NULL
#define fpga_resume NULL

static struct pci_driver fpga_driver = {
	.name = "fpga_driver",
	.id_table = fpga_id,
	.probe = fpga_probe,
	.remove = fpga_remove,
	.suspend = fpga_suspend,
	.resume = fpga_resume,
};

static int fpga_init(void)
{
	int ret;

	ret = pci_register_driver(&fpga_driver);
	if (ret) {
		pr_err("pci_register_driver() failed for fpga device\n");
		return ret;
	}

	pr_info("FPGA driver loaded\n");
	return 0;
}

static void fpga_exit(void)
{
	pci_unregister_driver(&fpga_driver);
	pr_info("FPGA driver unloaded\n");
}

/*-----------------------------------------------------------------------------
 *
 * Module init/exit
 */

static int __init dellemc_s5232f_init(void)
{
	int ret = 0;

	ret = i2c_init();
	if (ret) {
		pr_err("I2C subsystem initialization failed\n");
		return ret;
	}

	ret = fpga_init();
	if (ret) {
		pr_err("FPGA initialization failed\n");
		i2c_exit();
		return ret;
	}

	pr_info(DRIVER_NAME ": version " DRIVER_VERSION
		" successfully loaded\n");
	return 0;
}

static void __exit dellemc_s5232f_exit(void)
{
	fpga_exit();
	i2c_exit();
	pr_info(DRIVER_NAME " driver successfully unloaded\n");
}

module_init(dellemc_s5232f_init);
module_exit(dellemc_s5232f_exit);

MODULE_AUTHOR("David Yen (dhyen@cumulusnetworks.com)");
MODULE_DESCRIPTION("Dell EMC S5232F Platform Support");
MODULE_VERSION(DRIVER_VERSION);
MODULE_LICENSE("GPL");
